@@ -22,6 +22,7 @@
 #define ZEND_HASH_H
 
 #include "zend.h"
+#include "zend_sort.h"
 
 #define HASH_KEY_IS_STRING 1
 #define HASH_KEY_IS_LONG 2
@@ -170,6 +171,7 @@ ZEND_API zend_result ZEND_FASTCALL zend_hash_str_del(HashTable *ht, const char *
 ZEND_API zend_result ZEND_FASTCALL zend_hash_str_del_ind(HashTable *ht, const char *key, size_t len);
 ZEND_API zend_result ZEND_FASTCALL zend_hash_index_del(HashTable *ht, zend_ulong h);
 ZEND_API void ZEND_FASTCALL zend_hash_del_bucket(HashTable *ht, Bucket *p);
+ZEND_API void ZEND_FASTCALL zend_hash_packed_del_val(HashTable *ht, zval *zv);
 
 /* Data retrieval */
 ZEND_API zval* ZEND_FASTCALL zend_hash_find(const HashTable *ht, zend_string *key);
@@ -178,7 +180,7 @@ ZEND_API zval* ZEND_FASTCALL zend_hash_index_find(const HashTable *ht, zend_ulon
 ZEND_API zval* ZEND_FASTCALL _zend_hash_index_find(const HashTable *ht, zend_ulong h);
 
 /* The same as zend_hash_find(), but hash value of the key must be already calculated. */
-ZEND_API zval* ZEND_FASTCALL zend_hash_find_known_hash(const HashTable *ht, zend_string *key);
+ZEND_API zval* ZEND_FASTCALL zend_hash_find_known_hash(const HashTable *ht, const zend_string *key);
 
 static zend_always_inline zval *zend_hash_find_ex(const HashTable *ht, zend_string *key, bool known_hash)
 {
@@ -190,9 +192,9 @@ static zend_always_inline zval *zend_hash_find_ex(const HashTable *ht, zend_stri
 }
 
 #define ZEND_HASH_INDEX_FIND(_ht, _h, _ret, _not_found) do { \
-		if (EXPECTED(HT_FLAGS(_ht) & HASH_FLAG_PACKED)) { \
+		if (EXPECTED(HT_IS_PACKED(_ht))) { \
 			if (EXPECTED((zend_ulong)(_h) < (zend_ulong)(_ht)->nNumUsed)) { \
-				_ret = &_ht->arData[_h].val; \
+				_ret = &_ht->arPacked[_h]; \
 				if (UNEXPECTED(Z_TYPE_P(_ret) == IS_UNDEF)) { \
 					goto _not_found; \
 				} \
@@ -213,9 +215,9 @@ ZEND_API zval* ZEND_FASTCALL zend_hash_lookup(HashTable *ht, zend_string *key);
 ZEND_API zval* ZEND_FASTCALL zend_hash_index_lookup(HashTable *ht, zend_ulong h);
 
 #define ZEND_HASH_INDEX_LOOKUP(_ht, _h, _ret) do { \
-		if (EXPECTED(HT_FLAGS(_ht) & HASH_FLAG_PACKED)) { \
+		if (EXPECTED(HT_IS_PACKED(_ht))) { \
 			if (EXPECTED((zend_ulong)(_h) < (zend_ulong)(_ht)->nNumUsed)) { \
-				_ret = &_ht->arData[_h].val; \
+				_ret = &_ht->arPacked[_h]; \
 				if (EXPECTED(Z_TYPE_P(_ret) != IS_UNDEF)) { \
 					break; \
 				} \
@@ -243,35 +245,45 @@ static zend_always_inline bool zend_hash_index_exists(const HashTable *ht, zend_
 /* traversing */
 ZEND_API HashPosition ZEND_FASTCALL zend_hash_get_current_pos(const HashTable *ht);
 
-#define zend_hash_has_more_elements_ex(ht, pos) \
-	(zend_hash_get_current_key_type_ex(ht, pos) == HASH_KEY_NON_EXISTENT ? FAILURE : SUCCESS)
 ZEND_API zend_result   ZEND_FASTCALL zend_hash_move_forward_ex(HashTable *ht, HashPosition *pos);
 ZEND_API zend_result   ZEND_FASTCALL zend_hash_move_backwards_ex(HashTable *ht, HashPosition *pos);
-ZEND_API int   ZEND_FASTCALL zend_hash_get_current_key_ex(const HashTable *ht, zend_string **str_index, zend_ulong *num_index, HashPosition *pos);
-ZEND_API void  ZEND_FASTCALL zend_hash_get_current_key_zval_ex(const HashTable *ht, zval *key, HashPosition *pos);
+ZEND_API int   ZEND_FASTCALL zend_hash_get_current_key_ex(const HashTable *ht, zend_string **str_index, zend_ulong *num_index, const HashPosition *pos);
+ZEND_API void  ZEND_FASTCALL zend_hash_get_current_key_zval_ex(const HashTable *ht, zval *key, const HashPosition *pos);
 ZEND_API int   ZEND_FASTCALL zend_hash_get_current_key_type_ex(HashTable *ht, HashPosition *pos);
 ZEND_API zval* ZEND_FASTCALL zend_hash_get_current_data_ex(HashTable *ht, HashPosition *pos);
 ZEND_API void  ZEND_FASTCALL zend_hash_internal_pointer_reset_ex(HashTable *ht, HashPosition *pos);
 ZEND_API void  ZEND_FASTCALL zend_hash_internal_pointer_end_ex(HashTable *ht, HashPosition *pos);
 
-#define zend_hash_has_more_elements(ht) \
-	zend_hash_has_more_elements_ex(ht, &(ht)->nInternalPointer)
-#define zend_hash_move_forward(ht) \
-	zend_hash_move_forward_ex(ht, &(ht)->nInternalPointer)
-#define zend_hash_move_backwards(ht) \
-	zend_hash_move_backwards_ex(ht, &(ht)->nInternalPointer)
-#define zend_hash_get_current_key(ht, str_index, num_index) \
-	zend_hash_get_current_key_ex(ht, str_index, num_index, &(ht)->nInternalPointer)
-#define zend_hash_get_current_key_zval(ht, key) \
-	zend_hash_get_current_key_zval_ex(ht, key, &(ht)->nInternalPointer)
-#define zend_hash_get_current_key_type(ht) \
-	zend_hash_get_current_key_type_ex(ht, &(ht)->nInternalPointer)
-#define zend_hash_get_current_data(ht) \
-	zend_hash_get_current_data_ex(ht, &(ht)->nInternalPointer)
-#define zend_hash_internal_pointer_reset(ht) \
-	zend_hash_internal_pointer_reset_ex(ht, &(ht)->nInternalPointer)
-#define zend_hash_internal_pointer_end(ht) \
-	zend_hash_internal_pointer_end_ex(ht, &(ht)->nInternalPointer)
+static zend_always_inline zend_result zend_hash_has_more_elements_ex(HashTable *ht, HashPosition *pos) {
+	return (zend_hash_get_current_key_type_ex(ht, pos) == HASH_KEY_NON_EXISTENT ? FAILURE : SUCCESS);
+}
+static zend_always_inline zend_result zend_hash_has_more_elements(HashTable *ht) {
+	return zend_hash_has_more_elements_ex(ht, &ht->nInternalPointer);
+}
+static zend_always_inline zend_result zend_hash_move_forward(HashTable *ht) {
+	return zend_hash_move_forward_ex(ht, &ht->nInternalPointer);
+}
+static zend_always_inline zend_result zend_hash_move_backwards(HashTable *ht) {
+	return zend_hash_move_backwards_ex(ht, &ht->nInternalPointer);
+}
+static zend_always_inline int zend_hash_get_current_key(const HashTable *ht, zend_string **str_index, zend_ulong *num_index) {
+	return zend_hash_get_current_key_ex(ht, str_index, num_index, &ht->nInternalPointer);
+}
+static zend_always_inline void zend_hash_get_current_key_zval(const HashTable *ht, zval *key) {
+	zend_hash_get_current_key_zval_ex(ht, key, &ht->nInternalPointer);
+}
+static zend_always_inline int zend_hash_get_current_key_type(HashTable *ht) {
+	return zend_hash_get_current_key_type_ex(ht, &ht->nInternalPointer);
+}
+static zend_always_inline zval* zend_hash_get_current_data(HashTable *ht) {
+	return zend_hash_get_current_data_ex(ht, &ht->nInternalPointer);
+}
+static zend_always_inline void zend_hash_internal_pointer_reset(HashTable *ht) {
+	zend_hash_internal_pointer_reset_ex(ht, &ht->nInternalPointer);
+}
+static zend_always_inline void zend_hash_internal_pointer_end(HashTable *ht) {
+	zend_hash_internal_pointer_end_ex(ht, &ht->nInternalPointer);
+}
 
 /* Copying, merging and sorting */
 ZEND_API void  ZEND_FASTCALL zend_hash_copy(HashTable *target, HashTable *source, copy_ctor_func_t pCopyConstructor);
@@ -284,16 +296,19 @@ ZEND_API void  zend_hash_bucket_packed_swap(Bucket *p, Bucket *q);
 typedef int (*bucket_compare_func_t)(Bucket *a, Bucket *b);
 ZEND_API int   zend_hash_compare(HashTable *ht1, HashTable *ht2, compare_func_t compar, bool ordered);
 ZEND_API void  ZEND_FASTCALL zend_hash_sort_ex(HashTable *ht, sort_func_t sort_func, bucket_compare_func_t compare_func, bool renumber);
-ZEND_API zval* ZEND_FASTCALL zend_hash_minmax(const HashTable *ht, bucket_compare_func_t compar, uint32_t flag);
+ZEND_API zval* ZEND_FASTCALL zend_hash_minmax(const HashTable *ht, compare_func_t compar, uint32_t flag);
 
-#define zend_hash_sort(ht, compare_func, renumber) \
-	zend_hash_sort_ex(ht, zend_sort, compare_func, renumber)
+static zend_always_inline void ZEND_FASTCALL zend_hash_sort(HashTable *ht, bucket_compare_func_t compare_func, bool renumber) {
+	zend_hash_sort_ex(ht, zend_sort, compare_func, renumber);
+}
 
-#define zend_hash_num_elements(ht) \
-	(ht)->nNumOfElements
+static zend_always_inline uint32_t zend_hash_num_elements(const HashTable *ht) {
+	return ht->nNumOfElements;
+}
 
-#define zend_hash_next_free_element(ht) \
-	(ht)->nNextFreeElement
+static zend_always_inline zend_long zend_hash_next_free_element(const HashTable *ht) {
+	return ht->nNextFreeElement;
+}
 
 ZEND_API void ZEND_FASTCALL zend_hash_rehash(HashTable *ht);
 
@@ -319,6 +334,7 @@ ZEND_API HashTable* ZEND_FASTCALL zend_new_pair(zval *val1, zval *val2);
 ZEND_API uint32_t zend_array_count(HashTable *ht);
 ZEND_API HashTable* ZEND_FASTCALL zend_array_dup(HashTable *source);
 ZEND_API void ZEND_FASTCALL zend_array_destroy(HashTable *ht);
+ZEND_API HashTable* zend_array_to_list(HashTable *source);
 ZEND_API void ZEND_FASTCALL zend_symtable_clean(HashTable *ht);
 ZEND_API HashTable* ZEND_FASTCALL zend_symtable_to_proptable(HashTable *ht);
 ZEND_API HashTable* ZEND_FASTCALL zend_proptable_to_symtable(HashTable *ht, bool always_duplicate);
@@ -958,24 +974,281 @@ static zend_always_inline void *zend_hash_get_current_data_ptr_ex(HashTable *ht,
 #define zend_hash_get_current_data_ptr(ht) \
 	zend_hash_get_current_data_ptr_ex(ht, &(ht)->nInternalPointer)
 
+/* Common hash/packed array iterators */
+#if 0
+# define ZEND_HASH_ELEMENT_SIZE(__ht) \
+	(HT_IS_PACKED(__ht) ? sizeof(zval) : sizeof(Bucket))
+#else /* optimized version */
+# define ZEND_HASH_ELEMENT_SIZE(__ht) \
+	(sizeof(zval) + (~HT_FLAGS(__ht) & HASH_FLAG_PACKED) * ((sizeof(Bucket)-sizeof(zval))/HASH_FLAG_PACKED))
+#endif
+
+#define ZEND_HASH_ELEMENT_EX(__ht, _idx, _size) \
+	((zval*)(((char*)(__ht)->arPacked) + ((_idx) * (_size))))
+
+#define ZEND_HASH_ELEMENT(__ht, _idx) \
+	ZEND_HASH_ELEMENT_EX(__ht, _idx, ZEND_HASH_ELEMENT_SIZE(__ht))
+
+#define ZEND_HASH_NEXT_ELEMENT(_el, _size) \
+	((zval*)(((char*)(_el)) + (_size)))
+
+#define ZEND_HASH_PREV_ELEMENT(_el, _size) \
+	((zval*)(((char*)(_el)) - (_size)))
+
+#define _ZEND_HASH_FOREACH_VAL(_ht) do { \
+		HashTable *__ht = (_ht); \
+		uint32_t _count = __ht->nNumUsed; \
+		size_t _size = ZEND_HASH_ELEMENT_SIZE(__ht); \
+		zval *_z = __ht->arPacked; \
+		for (; _count > 0; _z = ZEND_HASH_NEXT_ELEMENT(_z, _size), _count--) { \
+			if (UNEXPECTED(Z_TYPE_P(_z) == IS_UNDEF)) continue;
+
+#define _ZEND_HASH_REVERSE_FOREACH_VAL(_ht) do { \
+		HashTable *__ht = (_ht); \
+		uint32_t _idx = __ht->nNumUsed; \
+		size_t _size = ZEND_HASH_ELEMENT_SIZE(__ht); \
+		zval *_z = ZEND_HASH_ELEMENT_EX(__ht, _idx, _size); \
+		for (;_idx > 0; _idx--) { \
+			_z = ZEND_HASH_PREV_ELEMENT(_z, _size); \
+			if (UNEXPECTED(Z_TYPE_P(_z) == IS_UNDEF)) continue;
+
 #define ZEND_HASH_FOREACH_FROM(_ht, indirect, _from) do { \
 		HashTable *__ht = (_ht); \
-		Bucket *_p = __ht->arData + (_from); \
-		Bucket *_end = __ht->arData + __ht->nNumUsed; \
-		for (; _p != _end; _p++) { \
-			zval *_z = &_p->val; \
-			if (indirect && Z_TYPE_P(_z) == IS_INDIRECT) { \
-				_z = Z_INDIRECT_P(_z); \
+		zend_ulong __h; \
+		zend_string *__key = NULL; \
+		uint32_t _idx = (_from); \
+		size_t _size = ZEND_HASH_ELEMENT_SIZE(__ht); \
+		zval *__z = ZEND_HASH_ELEMENT_EX(__ht, _idx, _size); \
+		uint32_t _count = __ht->nNumUsed - _idx; \
+		for (;_count > 0; _count--) { \
+			zval *_z = __z; \
+			if (HT_IS_PACKED(__ht)) { \
+				__z++; \
+				__h = _idx; \
+				_idx++; \
+			} else { \
+				Bucket *_p = (Bucket*)__z; \
+				__z = &(_p + 1)->val; \
+				__h = _p->h; \
+				__key = _p->key; \
+				if (indirect && Z_TYPE_P(_z) == IS_INDIRECT) { \
+					_z = Z_INDIRECT_P(_z); \
+				} \
 			} \
+			(void) __h; (void) __key; (void) _idx; \
 			if (UNEXPECTED(Z_TYPE_P(_z) == IS_UNDEF)) continue;
 
 #define ZEND_HASH_FOREACH(_ht, indirect) ZEND_HASH_FOREACH_FROM(_ht, indirect, 0)
 
 #define ZEND_HASH_REVERSE_FOREACH(_ht, indirect) do { \
+		HashTable *__ht = (_ht); \
+		uint32_t _idx = __ht->nNumUsed; \
+		zval *_z; \
+		zend_ulong __h; \
+		zend_string *__key = NULL; \
+		size_t _size = ZEND_HASH_ELEMENT_SIZE(__ht); \
+		zval *__z = ZEND_HASH_ELEMENT_EX(__ht, _idx, _size); \
+		for (;_idx > 0; _idx--) { \
+			if (HT_IS_PACKED(__ht)) { \
+				__z--; \
+				_z = __z; \
+				__h = _idx - 1; \
+			} else { \
+				Bucket *_p = (Bucket*)__z; \
+				_p--; \
+				__z = &_p->val; \
+				_z = __z; \
+				__h = _p->h; \
+				__key = _p->key; \
+				if (indirect && Z_TYPE_P(_z) == IS_INDIRECT) { \
+					_z = Z_INDIRECT_P(_z); \
+				} \
+			} \
+			(void) __h; (void) __key; (void) __z; \
+			if (UNEXPECTED(Z_TYPE_P(_z) == IS_UNDEF)) continue;
+
+#define ZEND_HASH_FOREACH_END() \
+		} \
+	} while (0)
+
+#define ZEND_HASH_FOREACH_END_DEL() \
+	ZEND_HASH_MAP_FOREACH_END_DEL()
+
+#define ZEND_HASH_FOREACH_BUCKET(ht, _bucket) \
+	ZEND_HASH_MAP_FOREACH_BUCKET(ht, _bucket)
+
+#define ZEND_HASH_FOREACH_BUCKET_FROM(ht, _bucket, _from) \
+	ZEND_HASH_MAP_FOREACH_BUCKET_FROM(ht, _bucket, _from)
+
+#define ZEND_HASH_REVERSE_FOREACH_BUCKET(ht, _bucket) \
+	ZEND_HASH_MAP_REVERSE_FOREACH_BUCKET(ht, _bucket)
+
+#define ZEND_HASH_FOREACH_VAL(ht, _val) \
+	_ZEND_HASH_FOREACH_VAL(ht); \
+	_val = _z;
+
+#define ZEND_HASH_REVERSE_FOREACH_VAL(ht, _val) \
+	_ZEND_HASH_REVERSE_FOREACH_VAL(ht); \
+	_val = _z;
+
+#define ZEND_HASH_FOREACH_VAL_IND(ht, _val) \
+	ZEND_HASH_FOREACH(ht, 1); \
+	_val = _z;
+
+#define ZEND_HASH_REVERSE_FOREACH_VAL_IND(ht, _val) \
+	ZEND_HASH_REVERSE_FOREACH(ht, 1); \
+	_val = _z;
+
+#define ZEND_HASH_FOREACH_PTR(ht, _ptr) \
+	_ZEND_HASH_FOREACH_VAL(ht); \
+	_ptr = Z_PTR_P(_z);
+
+#define ZEND_HASH_FOREACH_PTR_FROM(ht, _ptr, _from) \
+	ZEND_HASH_FOREACH_FROM(ht, 0, _from); \
+	_ptr = Z_PTR_P(_z);
+
+#define ZEND_HASH_REVERSE_FOREACH_PTR(ht, _ptr) \
+	_ZEND_HASH_REVERSE_FOREACH_VAL(ht); \
+	_ptr = Z_PTR_P(_z);
+
+#define ZEND_HASH_FOREACH_NUM_KEY(ht, _h) \
+	ZEND_HASH_FOREACH(ht, 0); \
+	_h = __h;
+
+#define ZEND_HASH_REVERSE_FOREACH_NUM_KEY(ht, _h) \
+	ZEND_HASH_REVERSE_FOREACH(ht, 0); \
+	_h = __h;
+
+#define ZEND_HASH_FOREACH_STR_KEY(ht, _key) \
+	ZEND_HASH_FOREACH(ht, 0); \
+	_key = __key;
+
+#define ZEND_HASH_REVERSE_FOREACH_STR_KEY(ht, _key) \
+	ZEND_HASH_REVERSE_FOREACH(ht, 0); \
+	_key = __key;
+
+#define ZEND_HASH_FOREACH_KEY(ht, _h, _key) \
+	ZEND_HASH_FOREACH(ht, 0); \
+	_h = __h; \
+	_key = __key;
+
+#define ZEND_HASH_REVERSE_FOREACH_KEY(ht, _h, _key) \
+	ZEND_HASH_REVERSE_FOREACH(ht, 0); \
+	_h = __h; \
+	_key = __key;
+
+#define ZEND_HASH_FOREACH_NUM_KEY_VAL(ht, _h, _val) \
+	ZEND_HASH_FOREACH(ht, 0); \
+	_h = __h; \
+	_val = _z;
+
+#define ZEND_HASH_REVERSE_FOREACH_NUM_KEY_VAL(ht, _h, _val) \
+	ZEND_HASH_REVERSE_FOREACH(ht, 0); \
+	_h = __h; \
+	_val = _z;
+
+#define ZEND_HASH_FOREACH_STR_KEY_VAL(ht, _key, _val) \
+	ZEND_HASH_FOREACH(ht, 0); \
+	_key = __key; \
+	_val = _z;
+
+#define ZEND_HASH_FOREACH_STR_KEY_VAL_FROM(ht, _key, _val, _from) \
+	ZEND_HASH_FOREACH_FROM(ht, 0, _from); \
+	_key = __key; \
+	_val = _z;
+
+#define ZEND_HASH_REVERSE_FOREACH_STR_KEY_VAL(ht, _key, _val) \
+	ZEND_HASH_REVERSE_FOREACH(ht, 0); \
+	_key = __key; \
+	_val = _z;
+
+#define ZEND_HASH_FOREACH_KEY_VAL(ht, _h, _key, _val) \
+	ZEND_HASH_FOREACH(ht, 0); \
+	_h = __h; \
+	_key = __key; \
+	_val = _z;
+
+#define ZEND_HASH_REVERSE_FOREACH_KEY_VAL(ht, _h, _key, _val) \
+	ZEND_HASH_REVERSE_FOREACH(ht, 0); \
+	_h = __h; \
+	_key = __key; \
+	_val = _z;
+
+#define ZEND_HASH_FOREACH_STR_KEY_VAL_IND(ht, _key, _val) \
+	ZEND_HASH_FOREACH(ht, 1); \
+	_key = __key; \
+	_val = _z;
+
+#define ZEND_HASH_REVERSE_FOREACH_STR_KEY_VAL_IND(ht, _key, _val) \
+	ZEND_HASH_REVERSE_FOREACH(ht, 1); \
+	_key = __key; \
+	_val = _z;
+
+#define ZEND_HASH_FOREACH_KEY_VAL_IND(ht, _h, _key, _val) \
+	ZEND_HASH_FOREACH(ht, 1); \
+	_h = __h; \
+	_key = __key; \
+	_val = _z;
+
+#define ZEND_HASH_REVERSE_FOREACH_KEY_VAL_IND(ht, _h, _key, _val) \
+	ZEND_HASH_REVERSE_FOREACH(ht, 1); \
+	_h = __h; \
+	_key = __key; \
+	_val = _z;
+
+#define ZEND_HASH_FOREACH_NUM_KEY_PTR(ht, _h, _ptr) \
+	ZEND_HASH_FOREACH(ht, 0); \
+	_h = __h; \
+	_ptr = Z_PTR_P(_z);
+
+#define ZEND_HASH_REVERSE_FOREACH_NUM_KEY_PTR(ht, _h, _ptr) \
+	ZEND_HASH_REVERSE_FOREACH(ht, 0); \
+	_h = __h; \
+	_ptr = Z_PTR_P(_z);
+
+#define ZEND_HASH_FOREACH_STR_KEY_PTR(ht, _key, _ptr) \
+	ZEND_HASH_FOREACH(ht, 0); \
+	_key = __key; \
+	_ptr = Z_PTR_P(_z);
+
+#define ZEND_HASH_REVERSE_FOREACH_STR_KEY_PTR(ht, _key, _ptr) \
+	ZEND_HASH_REVERSE_FOREACH(ht, 0); \
+	_key = __key; \
+	_ptr = Z_PTR_P(_z);
+
+#define ZEND_HASH_FOREACH_KEY_PTR(ht, _h, _key, _ptr) \
+	ZEND_HASH_FOREACH(ht, 0); \
+	_h = __h; \
+	_key = __key; \
+	_ptr = Z_PTR_P(_z);
+
+#define ZEND_HASH_REVERSE_FOREACH_KEY_PTR(ht, _h, _key, _ptr) \
+	ZEND_HASH_REVERSE_FOREACH(ht, 0); \
+	_h = __h; \
+	_key = __key; \
+	_ptr = Z_PTR_P(_z);
+
+/* Hash array iterators */
+#define ZEND_HASH_MAP_FOREACH_FROM(_ht, indirect, _from) do { \
+		HashTable *__ht = (_ht); \
+		Bucket *_p = __ht->arData + (_from); \
+		Bucket *_end = __ht->arData + __ht->nNumUsed; \
+		ZEND_ASSERT(!HT_IS_PACKED(__ht)); \
+		for (; _p != _end; _p++) { \
+			zval *_z = &_p->val; \
+			if (indirect && Z_TYPE_P(_z) == IS_INDIRECT) { \
+				_z = Z_INDIRECT_P(_z); \
+			} \
+			if (UNEXPECTED(Z_TYPE_P(_z) == IS_UNDEF)) continue;
+
+#define ZEND_HASH_MAP_FOREACH(_ht, indirect) ZEND_HASH_MAP_FOREACH_FROM(_ht, indirect, 0)
+
+#define ZEND_HASH_MAP_REVERSE_FOREACH(_ht, indirect) do { \
 		HashTable *__ht = (_ht); \
 		uint32_t _idx = __ht->nNumUsed; \
 		Bucket *_p = __ht->arData + _idx; \
 		zval *_z; \
+		ZEND_ASSERT(!HT_IS_PACKED(__ht)); \
 		for (_idx = __ht->nNumUsed; _idx > 0; _idx--) { \
 			_p--; \
 			_z = &_p->val; \
@@ -984,11 +1257,8 @@ static zend_always_inline void *zend_hash_get_current_data_ptr_ex(HashTable *ht,
 			} \
 			if (UNEXPECTED(Z_TYPE_P(_z) == IS_UNDEF)) continue;
 
-#define ZEND_HASH_FOREACH_END() \
-		} \
-	} while (0)
-
-#define ZEND_HASH_FOREACH_END_DEL() \
+#define ZEND_HASH_MAP_FOREACH_END_DEL() \
+			ZEND_ASSERT(!HT_IS_PACKED(__ht)); \
 			__ht->nNumOfElements--; \
 			do { \
 				uint32_t j = HT_IDX_TO_HASH(_idx - 1); \
@@ -1009,163 +1279,231 @@ static zend_always_inline void *zend_hash_get_current_data_ptr_ex(HashTable *ht,
 		__ht->nNumUsed = _idx; \
 	} while (0)
 
-#define ZEND_HASH_FOREACH_BUCKET(ht, _bucket) \
-	ZEND_HASH_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_FOREACH_BUCKET(ht, _bucket) \
+	ZEND_HASH_MAP_FOREACH(ht, 0); \
 	_bucket = _p;
 
-#define ZEND_HASH_FOREACH_BUCKET_FROM(ht, _bucket, _from) \
-	ZEND_HASH_FOREACH_FROM(ht, 0, _from); \
+#define ZEND_HASH_MAP_FOREACH_BUCKET_FROM(ht, _bucket, _from) \
+	ZEND_HASH_MAP_FOREACH_FROM(ht, 0, _from); \
 	_bucket = _p;
 
-#define ZEND_HASH_REVERSE_FOREACH_BUCKET(ht, _bucket) \
-	ZEND_HASH_REVERSE_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_REVERSE_FOREACH_BUCKET(ht, _bucket) \
+	ZEND_HASH_MAP_REVERSE_FOREACH(ht, 0); \
 	_bucket = _p;
 
-#define ZEND_HASH_FOREACH_VAL(ht, _val) \
-	ZEND_HASH_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_FOREACH_VAL(ht, _val) \
+	ZEND_HASH_MAP_FOREACH(ht, 0); \
 	_val = _z;
 
-#define ZEND_HASH_REVERSE_FOREACH_VAL(ht, _val) \
-	ZEND_HASH_REVERSE_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_REVERSE_FOREACH_VAL(ht, _val) \
+	ZEND_HASH_MAP_REVERSE_FOREACH(ht, 0); \
 	_val = _z;
 
-#define ZEND_HASH_FOREACH_VAL_IND(ht, _val) \
-	ZEND_HASH_FOREACH(ht, 1); \
+#define ZEND_HASH_MAP_FOREACH_VAL_IND(ht, _val) \
+	ZEND_HASH_MAP_FOREACH(ht, 1); \
 	_val = _z;
 
-#define ZEND_HASH_REVERSE_FOREACH_VAL_IND(ht, _val) \
-	ZEND_HASH_REVERSE_FOREACH(ht, 1); \
+#define ZEND_HASH_MAP_REVERSE_FOREACH_VAL_IND(ht, _val) \
+	ZEND_HASH_MAP_REVERSE_FOREACH(ht, 1); \
 	_val = _z;
 
-#define ZEND_HASH_FOREACH_PTR(ht, _ptr) \
-	ZEND_HASH_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_FOREACH_PTR(ht, _ptr) \
+	ZEND_HASH_MAP_FOREACH(ht, 0); \
 	_ptr = Z_PTR_P(_z);
 
-#define ZEND_HASH_FOREACH_PTR_FROM(ht, _ptr, _from) \
-	ZEND_HASH_FOREACH_FROM(ht, 0, _from); \
+#define ZEND_HASH_MAP_FOREACH_PTR_FROM(ht, _ptr, _from) \
+	ZEND_HASH_MAP_FOREACH_FROM(ht, 0, _from); \
 	_ptr = Z_PTR_P(_z);
 
-#define ZEND_HASH_REVERSE_FOREACH_PTR(ht, _ptr) \
-	ZEND_HASH_REVERSE_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_REVERSE_FOREACH_PTR(ht, _ptr) \
+	ZEND_HASH_MAP_REVERSE_FOREACH(ht, 0); \
 	_ptr = Z_PTR_P(_z);
 
-#define ZEND_HASH_FOREACH_NUM_KEY(ht, _h) \
-	ZEND_HASH_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_FOREACH_NUM_KEY(ht, _h) \
+	ZEND_HASH_MAP_FOREACH(ht, 0); \
 	_h = _p->h;
 
-#define ZEND_HASH_REVERSE_FOREACH_NUM_KEY(ht, _h) \
-	ZEND_HASH_REVERSE_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_REVERSE_FOREACH_NUM_KEY(ht, _h) \
+	ZEND_HASH_MAP_REVERSE_FOREACH(ht, 0); \
 	_h = _p->h;
 
-#define ZEND_HASH_FOREACH_STR_KEY(ht, _key) \
-	ZEND_HASH_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_FOREACH_STR_KEY(ht, _key) \
+	ZEND_HASH_MAP_FOREACH(ht, 0); \
 	_key = _p->key;
 
-#define ZEND_HASH_REVERSE_FOREACH_STR_KEY(ht, _key) \
-	ZEND_HASH_REVERSE_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_REVERSE_FOREACH_STR_KEY(ht, _key) \
+	ZEND_HASH_MAP_REVERSE_FOREACH(ht, 0); \
 	_key = _p->key;
 
-#define ZEND_HASH_FOREACH_KEY(ht, _h, _key) \
-	ZEND_HASH_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_FOREACH_KEY(ht, _h, _key) \
+	ZEND_HASH_MAP_FOREACH(ht, 0); \
 	_h = _p->h; \
 	_key = _p->key;
 
-#define ZEND_HASH_REVERSE_FOREACH_KEY(ht, _h, _key) \
-	ZEND_HASH_REVERSE_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_REVERSE_FOREACH_KEY(ht, _h, _key) \
+	ZEND_HASH_MAP_REVERSE_FOREACH(ht, 0); \
 	_h = _p->h; \
 	_key = _p->key;
 
-#define ZEND_HASH_FOREACH_NUM_KEY_VAL(ht, _h, _val) \
-	ZEND_HASH_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_FOREACH_NUM_KEY_VAL(ht, _h, _val) \
+	ZEND_HASH_MAP_FOREACH(ht, 0); \
 	_h = _p->h; \
 	_val = _z;
 
-#define ZEND_HASH_REVERSE_FOREACH_NUM_KEY_VAL(ht, _h, _val) \
-	ZEND_HASH_REVERSE_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_REVERSE_FOREACH_NUM_KEY_VAL(ht, _h, _val) \
+	ZEND_HASH_MAP_REVERSE_FOREACH(ht, 0); \
 	_h = _p->h; \
 	_val = _z;
 
-#define ZEND_HASH_FOREACH_STR_KEY_VAL(ht, _key, _val) \
-	ZEND_HASH_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_FOREACH_STR_KEY_VAL(ht, _key, _val) \
+	ZEND_HASH_MAP_FOREACH(ht, 0); \
 	_key = _p->key; \
 	_val = _z;
 
-#define ZEND_HASH_FOREACH_STR_KEY_VAL_FROM(ht, _key, _val, _from) \
-	ZEND_HASH_FOREACH_FROM(ht, 0, _from); \
+#define ZEND_HASH_MAP_FOREACH_STR_KEY_VAL_FROM(ht, _key, _val, _from) \
+	ZEND_HASH_MAP_FOREACH_FROM(ht, 0, _from); \
 	_key = _p->key; \
 	_val = _z;
 
-#define ZEND_HASH_REVERSE_FOREACH_STR_KEY_VAL(ht, _key, _val) \
-	ZEND_HASH_REVERSE_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_REVERSE_FOREACH_STR_KEY_VAL(ht, _key, _val) \
+	ZEND_HASH_MAP_REVERSE_FOREACH(ht, 0); \
 	_key = _p->key; \
 	_val = _z;
 
-#define ZEND_HASH_FOREACH_KEY_VAL(ht, _h, _key, _val) \
-	ZEND_HASH_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_FOREACH_KEY_VAL(ht, _h, _key, _val) \
+	ZEND_HASH_MAP_FOREACH(ht, 0); \
 	_h = _p->h; \
 	_key = _p->key; \
 	_val = _z;
 
-#define ZEND_HASH_REVERSE_FOREACH_KEY_VAL(ht, _h, _key, _val) \
-	ZEND_HASH_REVERSE_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_REVERSE_FOREACH_KEY_VAL(ht, _h, _key, _val) \
+	ZEND_HASH_MAP_REVERSE_FOREACH(ht, 0); \
 	_h = _p->h; \
 	_key = _p->key; \
 	_val = _z;
 
-#define ZEND_HASH_FOREACH_STR_KEY_VAL_IND(ht, _key, _val) \
-	ZEND_HASH_FOREACH(ht, 1); \
+#define ZEND_HASH_MAP_FOREACH_STR_KEY_VAL_IND(ht, _key, _val) \
+	ZEND_HASH_MAP_FOREACH(ht, 1); \
 	_key = _p->key; \
 	_val = _z;
 
-#define ZEND_HASH_REVERSE_FOREACH_STR_KEY_VAL_IND(ht, _key, _val) \
-	ZEND_HASH_REVERSE_FOREACH(ht, 1); \
+#define ZEND_HASH_MAP_REVERSE_FOREACH_STR_KEY_VAL_IND(ht, _key, _val) \
+	ZEND_HASH_MAP_REVERSE_FOREACH(ht, 1); \
 	_key = _p->key; \
 	_val = _z;
 
-#define ZEND_HASH_FOREACH_KEY_VAL_IND(ht, _h, _key, _val) \
-	ZEND_HASH_FOREACH(ht, 1); \
+#define ZEND_HASH_MAP_FOREACH_KEY_VAL_IND(ht, _h, _key, _val) \
+	ZEND_HASH_MAP_FOREACH(ht, 1); \
 	_h = _p->h; \
 	_key = _p->key; \
 	_val = _z;
 
-#define ZEND_HASH_REVERSE_FOREACH_KEY_VAL_IND(ht, _h, _key, _val) \
-	ZEND_HASH_REVERSE_FOREACH(ht, 1); \
+#define ZEND_HASH_MAP_REVERSE_FOREACH_KEY_VAL_IND(ht, _h, _key, _val) \
+	ZEND_HASH_MAP_REVERSE_FOREACH(ht, 1); \
 	_h = _p->h; \
 	_key = _p->key; \
 	_val = _z;
 
-#define ZEND_HASH_FOREACH_NUM_KEY_PTR(ht, _h, _ptr) \
-	ZEND_HASH_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_FOREACH_NUM_KEY_PTR(ht, _h, _ptr) \
+	ZEND_HASH_MAP_FOREACH(ht, 0); \
 	_h = _p->h; \
 	_ptr = Z_PTR_P(_z);
 
-#define ZEND_HASH_REVERSE_FOREACH_NUM_KEY_PTR(ht, _h, _ptr) \
-	ZEND_HASH_REVERSE_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_REVERSE_FOREACH_NUM_KEY_PTR(ht, _h, _ptr) \
+	ZEND_HASH_MAP_REVERSE_FOREACH(ht, 0); \
 	_h = _p->h; \
 	_ptr = Z_PTR_P(_z);
 
-#define ZEND_HASH_FOREACH_STR_KEY_PTR(ht, _key, _ptr) \
-	ZEND_HASH_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_FOREACH_STR_KEY_PTR(ht, _key, _ptr) \
+	ZEND_HASH_MAP_FOREACH(ht, 0); \
 	_key = _p->key; \
 	_ptr = Z_PTR_P(_z);
 
-#define ZEND_HASH_REVERSE_FOREACH_STR_KEY_PTR(ht, _key, _ptr) \
-	ZEND_HASH_REVERSE_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_REVERSE_FOREACH_STR_KEY_PTR(ht, _key, _ptr) \
+	ZEND_HASH_MAP_REVERSE_FOREACH(ht, 0); \
 	_key = _p->key; \
 	_ptr = Z_PTR_P(_z);
 
-#define ZEND_HASH_FOREACH_KEY_PTR(ht, _h, _key, _ptr) \
-	ZEND_HASH_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_FOREACH_KEY_PTR(ht, _h, _key, _ptr) \
+	ZEND_HASH_MAP_FOREACH(ht, 0); \
 	_h = _p->h; \
 	_key = _p->key; \
 	_ptr = Z_PTR_P(_z);
 
-#define ZEND_HASH_REVERSE_FOREACH_KEY_PTR(ht, _h, _key, _ptr) \
-	ZEND_HASH_REVERSE_FOREACH(ht, 0); \
+#define ZEND_HASH_MAP_REVERSE_FOREACH_KEY_PTR(ht, _h, _key, _ptr) \
+	ZEND_HASH_MAP_REVERSE_FOREACH(ht, 0); \
 	_h = _p->h; \
 	_key = _p->key; \
 	_ptr = Z_PTR_P(_z);
 
+/* Packed array iterators */
+#define ZEND_HASH_PACKED_FOREACH_FROM(_ht, _from) do { \
+		HashTable *__ht = (_ht); \
+		zend_ulong _idx = (_from); \
+		zval *_z = __ht->arPacked + (_from); \
+		zval *_end = __ht->arPacked + __ht->nNumUsed; \
+		ZEND_ASSERT(HT_IS_PACKED(__ht)); \
+		for (;_z != _end; _z++, _idx++) { \
+			(void) _idx; \
+			if (UNEXPECTED(Z_TYPE_P(_z) == IS_UNDEF)) continue;
+
+#define ZEND_HASH_PACKED_FOREACH(_ht) ZEND_HASH_PACKED_FOREACH_FROM(_ht, 0)
+
+#define ZEND_HASH_PACKED_REVERSE_FOREACH(_ht) do { \
+		HashTable *__ht = (_ht); \
+		zend_ulong _idx = __ht->nNumUsed; \
+		zval *_z = __ht->arPacked + _idx; \
+		ZEND_ASSERT(HT_IS_PACKED(__ht)); \
+		while (_idx > 0) { \
+			_z--; \
+			_idx--; \
+			(void) _idx; \
+			if (UNEXPECTED(Z_TYPE_P(_z) == IS_UNDEF)) continue;
+
+#define ZEND_HASH_PACKED_FOREACH_VAL(ht, _val) \
+	ZEND_HASH_PACKED_FOREACH(ht); \
+	_val = _z;
+
+#define ZEND_HASH_PACKED_REVERSE_FOREACH_VAL(ht, _val) \
+	ZEND_HASH_PACKED_REVERSE_FOREACH(ht); \
+	_val = _z;
+
+#define ZEND_HASH_PACKED_FOREACH_PTR(ht, _ptr) \
+	ZEND_HASH_PACKED_FOREACH(ht); \
+	_ptr = Z_PTR_P(_z);
+
+#define ZEND_HASH_PACKED_REVERSE_FOREACH_PTR(ht, _ptr) \
+	ZEND_HASH_PACKED_REVERSE_FOREACH(ht); \
+	_ptr = Z_PTR_P(_z);
+
+#define ZEND_HASH_PACKED_FOREACH_KEY(ht, _h) \
+	ZEND_HASH_PACKED_FOREACH(ht); \
+	_h = _idx;
+
+#define ZEND_HASH_PACKED_REVERSE_FOREACH_KEY(ht, _h) \
+	ZEND_HASH_PACKED_REVERSE_FOREACH(ht); \
+	_h = _idx;
+
+#define ZEND_HASH_PACKED_FOREACH_KEY_VAL(ht, _h, _val) \
+	ZEND_HASH_PACKED_FOREACH(ht); \
+	_h = _idx; \
+	_val = _z;
+
+#define ZEND_HASH_PACKED_REVERSE_FOREACH_KEY_VAL(ht, _h, _val) \
+	ZEND_HASH_PACKED_REVERSE_FOREACH(ht); \
+	_h = _idx; \
+	_val = _z;
+
+#define ZEND_HASH_PACKED_FOREACH_KEY_PTR(ht, _h, _ptr) \
+	ZEND_HASH_PACKED_FOREACH(ht); \
+	_h = _idx; \
+	_ptr = Z_PTR_P(_z);
+
+#define ZEND_HASH_PACKED_REVERSE_FOREACH_KEY_PTR(ht, _h, _ptr) \
+	ZEND_HASH_PACKED_REVERSE_FOREACH(ht); \
+	_h = _idx; \
+	_ptr = Z_PTR_P(_z);
+
 /* The following macros are useful to insert a sequence of new elements
  * of packed array. They may be used instead of series of
  * zend_hash_next_index_insert_new()
@@ -1173,45 +1511,43 @@ static zend_always_inline void *zend_hash_get_current_data_ptr_ex(HashTable *ht,
  */
 #define ZEND_HASH_FILL_PACKED(ht) do { \
 		HashTable *__fill_ht = (ht); \
-		Bucket *__fill_bkt = __fill_ht->arData + __fill_ht->nNumUsed; \
+		zval *__fill_val = __fill_ht->arPacked + __fill_ht->nNumUsed; \
 		uint32_t __fill_idx = __fill_ht->nNumUsed; \
-		ZEND_ASSERT(HT_FLAGS(__fill_ht) & HASH_FLAG_PACKED);
+		ZEND_ASSERT(HT_IS_PACKED(__fill_ht));
 
 #define ZEND_HASH_FILL_GROW() do { \
 		if (UNEXPECTED(__fill_idx >= __fill_ht->nTableSize)) { \
+			__fill_ht->nNumOfElements += __fill_idx - __fill_ht->nNumUsed; \
 			__fill_ht->nNumUsed = __fill_idx; \
-			__fill_ht->nNumOfElements = __fill_idx; \
 			__fill_ht->nNextFreeElement = __fill_idx; \
 			zend_hash_packed_grow(__fill_ht); \
-			__fill_bkt = __fill_ht->arData + __fill_idx; \
+			__fill_val = __fill_ht->arPacked + __fill_idx; \
 		} \
 	} while (0);
 
 #define ZEND_HASH_FILL_SET(_val) \
-		ZVAL_COPY_VALUE(&__fill_bkt->val, _val)
+		ZVAL_COPY_VALUE(__fill_val, _val)
 
 #define ZEND_HASH_FILL_SET_NULL() \
-		ZVAL_NULL(&__fill_bkt->val)
+		ZVAL_NULL(__fill_val)
 
 #define ZEND_HASH_FILL_SET_LONG(_val) \
-		ZVAL_LONG(&__fill_bkt->val, _val)
+		ZVAL_LONG(__fill_val, _val)
 
 #define ZEND_HASH_FILL_SET_DOUBLE(_val) \
-		ZVAL_DOUBLE(&__fill_bkt->val, _val)
+		ZVAL_DOUBLE(__fill_val, _val)
 
 #define ZEND_HASH_FILL_SET_STR(_val) \
-		ZVAL_STR(&__fill_bkt->val, _val)
+		ZVAL_STR(__fill_val, _val)
 
 #define ZEND_HASH_FILL_SET_STR_COPY(_val) \
-		ZVAL_STR_COPY(&__fill_bkt->val, _val)
+		ZVAL_STR_COPY(__fill_val, _val)
 
 #define ZEND_HASH_FILL_SET_INTERNED_STR(_val) \
-		ZVAL_INTERNED_STR(&__fill_bkt->val, _val)
+		ZVAL_INTERNED_STR(__fill_val, _val)
 
 #define ZEND_HASH_FILL_NEXT() do {\
-		__fill_bkt->h = (__fill_idx); \
-		__fill_bkt->key = NULL; \
-		__fill_bkt++; \
+		__fill_val++; \
 		__fill_idx++; \
 	} while (0)
 
@@ -1221,8 +1557,8 @@ static zend_always_inline void *zend_hash_get_current_data_ptr_ex(HashTable *ht,
 	} while (0)
 
 #define ZEND_HASH_FILL_FINISH() do { \
+		__fill_ht->nNumOfElements += __fill_idx - __fill_ht->nNumUsed; \
 		__fill_ht->nNumUsed = __fill_idx; \
-		__fill_ht->nNumOfElements = __fill_idx; \
 		__fill_ht->nNextFreeElement = __fill_idx; \
 		__fill_ht->nInternalPointer = 0; \
 	} while (0)
@@ -1234,8 +1570,8 @@ static zend_always_inline void *zend_hash_get_current_data_ptr_ex(HashTable *ht,
 /* Check if an array is a list */
 static zend_always_inline bool zend_array_is_list(zend_array *array)
 {
-	zend_long expected_idx = 0;
-	zend_long num_idx;
+	zend_ulong expected_idx = 0;
+	zend_ulong num_idx;
 	zend_string* str_idx;
 	/* Empty arrays are lists */
 	if (zend_hash_num_elements(array) == 0) {
@@ -1243,16 +1579,24 @@ static zend_always_inline bool zend_array_is_list(zend_array *array)
 	}
 
 	/* Packed arrays are lists */
-	if (HT_IS_PACKED(array) && HT_IS_WITHOUT_HOLES(array)) {
-		return 1;
-	}
-
-	/* Check if the list could theoretically be repacked */
-	ZEND_HASH_FOREACH_KEY(array, num_idx, str_idx) {
-		if (str_idx != NULL || num_idx != expected_idx++) {
-			return 0;
+	if (HT_IS_PACKED(array)) {
+		if (HT_IS_WITHOUT_HOLES(array)) {
+			return 1;
 		}
-	} ZEND_HASH_FOREACH_END();
+		/* Check if the list could theoretically be repacked */
+		ZEND_HASH_PACKED_FOREACH_KEY(array, num_idx) {
+			if (num_idx != expected_idx++) {
+				return 0;
+			}
+		} ZEND_HASH_FOREACH_END();
+	} else {
+		/* Check if the list could theoretically be repacked */
+		ZEND_HASH_MAP_FOREACH_KEY(array, num_idx, str_idx) {
+			if (str_idx != NULL || num_idx != expected_idx++) {
+				return 0;
+			}
+		} ZEND_HASH_FOREACH_END();
+	}
 
 	return 1;
 }
